import { Webview, ExtensionContext, WebviewOptions, Uri, ViewColumn, WebviewPanel, WebviewPanelOptions, WebviewViewProvider, WebviewView } from 'vscode';
import { AsyncStateful, AsyncState, WebviewStreamMessage } from '@stack-spot/vscode-async-webview-shared';
export { AsyncState, AsyncStateful, ManualPromise, errorToString } from '@stack-spot/vscode-async-webview-shared';

/**
 * This makes the bridge between the extension and webview. Every public method declared here will be callable by both ends of the
 * application.
 *
 * The bridge also declares a State that lives in the webview, but can accessible by both parts. The Generics received by this class is
 * the type of this state.
 *
 * Every method declared here will be asynchronous when accessed from the webview.
 *
 * Every state declared here will be asynchronous when accessed by the extension, but synchronous in the webview.
 *
 * To access a method from the webview, `VSCodeWebInterface#bridge` must be used.
 * To access a state from the webview, `VSCodeWebInterface#getState`, `VSCodeWebInterface#initializeState`, `VSCodeWebInterface#setState`
 * and `VSCodeWebInterface#observeState` must be used. If you're using React, state management can be replaced by a simple hook.
 *
 * To access a state from the bridge (this class), use `await this.state.get(name)` and `await this.state.set(name, value)`. Tip: if you
 * don't need to know if the state has finished being set or not, you can omit the `await` to make the code more simple.
 *
 * If you need to do additional tasks before disposing the bridge, you can overwrite the method `dispose`, just don't forget to call
 * `super.dispose()`.
 */
declare abstract class VSCodeWebviewBridge<StateType extends Record<string, any> = Record<string, any>> implements AsyncStateful<StateType> {
    #private;
    readonly state: AsyncState<StateType>;
    constructor(webview: Webview);
    /**
     * Streams a string from the extension to the webview.
     * @param message the package to send.
     */
    stream(message: Omit<WebviewStreamMessage, 'type' | 'index'>): void;
    dispose(): void;
}

interface ViewOptions<Bridge extends VSCodeWebviewBridge> {
    /**
     * The path to the webview directory.
     * Attention:
     * - This must be the compiled/bundled version of the we application.
     * - This path must be relative to the root package.json in the final extension bundle.
     */
    path: string;
    /**
     * The name of the html file under `path` to show.
     * @default index.html
     */
    index?: string;
    /**
     * The title of the webview
     */
    title: string;
    /**
     * The ExtensionContext
     */
    context: ExtensionContext;
    /**
     * The type. This is the same `type` used by any VSCode view.
     */
    type?: string;
    /**
     * The options for the Webview.
     *
     * `enableScripts` and `localResourceRoots` will be set to the correct values unless explicitly overwritten by this configuration.
     */
    options?: WebviewOptions;
    /**
     * A factory that creates the Bridge between the extension and the webview. The Bridge must extend VSCodeWebviewBridge which requires a
     * Webview in its constructor. For this reason, this factory function receives a `Webview` as parameter.
     *
     * If this is not provided, the library won't be able to handle any communication between the two applications.
     *
     * @param webview the webview loaded
     * @returns an instance of the Bridge
     */
    bridgeFactory?: (webview: Webview) => Bridge;
}

interface ViewColumnWithFocus {
    viewColumn: ViewColumn;
    preserveFocus?: boolean;
}
type ShowOptions = ViewColumn | ViewColumnWithFocus;
type PanelOptions = WebviewPanelOptions & WebviewOptions;
interface Options<Bridge extends VSCodeWebviewBridge> extends ViewOptions<Bridge> {
    /**
     * Which ViewColumn would you like to use?
     */
    showOptions?: ShowOptions;
    /**
     * The options for the panel containing the webview.
     */
    options?: PanelOptions;
}
/**
 * Represents a Webview loaded into the extension. It creates a Bridge and a State shared between both the extension and the webview.
 *
 * It creates a Panel with a Webview. The panel is first created when `show()` is called for the first time.
 *
 * The Webview must be a separate project already compiled (bundled). `path`, in the constructor, must refer to the path to the site
 * directory inside the final extension bundle. This path must be relative to the package.json in the root of the bundle.
 *
 * If the file to load in the webview is not `index.html`, you should provide a value for `index` in the constructor.
 *
 * To show a panel with a webview, first create the `VSCodeWebview` with the required config and then call `VSCodeWebview#show()`.
 */
declare class VSCodeWebview<Bridge extends VSCodeWebviewBridge = VSCodeWebviewBridge> {
    protected readonly baseUri: Uri;
    private readonly title;
    readonly type: string;
    private readonly showOptions;
    protected readonly options: PanelOptions | undefined;
    private panel;
    protected readonly bridgeFactory: Options<Bridge>['bridgeFactory'];
    protected bridge: Bridge | undefined;
    private htmlPromise;
    private html;
    private readonly index;
    private bridgePromise;
    constructor({ path, index, title, type, showOptions, context, bridgeFactory, options, }: Options<Bridge>);
    protected buildHtml(baseSrc: Uri): Promise<string>;
    protected buildBridge(webview: Webview): void;
    private buildPanel;
    protected treatHTML(html: string, baseSrc: Uri): string;
    /**
     * Shows the webview panel.
     * If the panel has not yet been created, it's created.
     * @param column the column to show the panel at.
     */
    show(column?: ViewColumn): Promise<void>;
    /**
     * Gets a reference to the panel.
     *
     * If the panel has not yet been created or if it has been disposed, undefined is returned.
     *
     * @returns the current panel.
     */
    getPanel(): WebviewPanel | undefined;
    /**
     * Gets the HTML of the webview.
     *
     * If the webview hasn't been shown yet, this will be undefined.
     *
     * @returns the html.
     */
    getHTML(): string | undefined;
    /**
     * Gets a reference to the current bridge.
     *
     * If the bridge has not yet been created or it has been disposed, undefined is returned. The bridge is created/disposed at the same time
     * as the panel.
     *
     * @returns the current bridge.
     */
    getBridge(): Bridge | undefined;
    /**
     * Same as getBridge, but if the bridge has not yet been created, the promise returned will resolve once it's available.
     * @returns a promise that resolves to the bridge.
     */
    getBridgeAsync(): Promise<Bridge>;
}

/**
 * Same as {@link VSCodeWebview}, but instead of creating a new Panel, it implements the interface {@link WebviewViewProvider} instead.
 */
declare class VSCodeViewProvider<Bridge extends VSCodeWebviewBridge = VSCodeWebviewBridge> extends VSCodeWebview<Bridge> implements WebviewViewProvider {
    private view;
    constructor(options: ViewOptions<Bridge>);
    resolveWebviewView(view: WebviewView): Promise<void>;
    /**
     * Shows the current view if it's hidden.
     * Same as running the command "{viewType}.focus".
     *
     * If the view is not available, an error message is shown as a notification in VSCode.
     */
    show(): Promise<void>;
    /**
     * Gets the current WebviewView.
     *
     * If the vire has not yet been created, undefined is returned.
     *
     * @returns the current view.
     */
    getView(): WebviewView | undefined;
}

export { VSCodeViewProvider, VSCodeWebview, VSCodeWebviewBridge };
