import { readFile } from 'fs/promises';
import { camelCase } from 'lodash';
import { Uri, ViewColumn, window } from 'vscode';
import { errorToString, ManualPromise, logger, buildBridgeError, buildBridgeResponse, messageType, buildGetStateRequest, buildSetStateRequest, asWebViewMessage } from '@stack-spot/vscode-async-webview-shared';
export { AsyncState, AsyncStateful, ManualPromise, errorToString } from '@stack-spot/vscode-async-webview-shared';

class VSCodeWebview {
  baseUri;
  title;
  type;
  showOptions;
  options;
  panel;
  bridgeFactory;
  bridge;
  htmlPromise;
  html;
  index;
  bridgePromise;
  constructor({
    path,
    index = "index.html",
    title,
    type = camelCase(title),
    showOptions = ViewColumn.One,
    context,
    bridgeFactory,
    options
  }) {
    this.baseUri = Uri.joinPath(context.extensionUri, path);
    this.title = title;
    this.type = type;
    this.showOptions = showOptions;
    this.bridgeFactory = bridgeFactory;
    const basePath = process.platform === "win32" ? this.baseUri.path.replace(/^\//, "") : this.baseUri.path;
    this.htmlPromise = readFile(`${basePath}/${index}`, { encoding: "utf-8" });
    this.index = index;
    this.options = {
      enableScripts: true,
      localResourceRoots: [this.baseUri],
      ...options
    };
  }
  async buildHtml(baseSrc) {
    try {
      const htmlText = await this.htmlPromise;
      this.html = this.treatHTML(htmlText, baseSrc);
    } catch (error) {
      window.showErrorMessage("There was an error while loading the html for the webview. This is a bug, please report it to the team.");
      this.html = `
        <html>
          <body>
            <p>Unable to load webview from ${this.baseUri.path}/${this.index}.</p>
            <p>The underlying error is: ${errorToString(error)}</p>
          </body>
        </html>
      `;
    }
    return this.html;
  }
  buildBridge(webview) {
    try {
      this.bridge = this.bridgeFactory?.call(null, webview);
      if (this.bridge && this.bridgePromise) {
        this.bridgePromise.resolve(this.bridge);
        this.bridgePromise = void 0;
      }
    } catch (error) {
      window.showErrorMessage([
        "There was an error while building the webview: unable to instantiate the webview's Bridge.",
        `This is a bug, please report it to the team. Cause: ${errorToString(error)}`
      ].join("\n"));
    }
  }
  async buildPanel() {
    this.panel = window.createWebviewPanel(this.type, this.title, this.showOptions, this.options);
    this.buildBridge(this.panel.webview);
    const html = await this.buildHtml(this.panel?.webview.asWebviewUri(this.baseUri));
    this.panel.webview.html = html;
    this.panel.onDidDispose(() => {
      this.panel = void 0;
      this.bridge?.dispose();
      this.bridge = void 0;
    });
  }
  treatHTML(html, baseSrc) {
    return html.replace("<head>", `<head><base href="${baseSrc}/">`);
  }
  /**
   * Shows the webview panel.
   * If the panel has not yet been created, it's created.
   * @param column the column to show the panel at.
   */
  async show(column) {
    if (this.panel)
      this.panel.reveal(column);
    else
      await this.buildPanel();
  }
  /**
   * Gets a reference to the panel.
   * 
   * If the panel has not yet been created or if it has been disposed, undefined is returned.
   * 
   * @returns the current panel.
   */
  getPanel() {
    return this.panel;
  }
  /**
   * Gets the HTML of the webview.
   * 
   * If the webview hasn't been shown yet, this will be undefined.
   * 
   * @returns the html.
   */
  getHTML() {
    return this.html;
  }
  /**
   * Gets a reference to the current bridge.
   * 
   * If the bridge has not yet been created or it has been disposed, undefined is returned. The bridge is created/disposed at the same time
   * as the panel.
   * 
   * @returns the current bridge.
   */
  getBridge() {
    return this.bridge;
  }
  /**
   * Same as getBridge, but if the bridge has not yet been created, the promise returned will resolve once it's available.
   * @returns a promise that resolves to the bridge.
   */
  getBridgeAsync() {
    if (this.bridge)
      return Promise.resolve(this.bridge);
    this.bridgePromise = new ManualPromise();
    return this.bridgePromise.promise;
  }
}

class VSCodeViewProvider extends VSCodeWebview {
  view;
  constructor(options) {
    super(options);
    if (options.type) {
      options.context.subscriptions.push(window.registerWebviewViewProvider(options.type, this));
    }
  }
  async resolveWebviewView(view) {
    this.view = view;
    const { webview } = view;
    const html = this.getHTML() ?? await this.buildHtml(webview.asWebviewUri(this.baseUri));
    this.buildBridge(webview);
    webview.options = this.options;
    webview.html = html;
    view.onDidDispose(() => {
      this.view = void 0;
      this.bridge?.dispose();
      this.bridge = void 0;
    });
  }
  /**
   * Shows the current view if it's hidden.
   * Same as running the command "{viewType}.focus".
   * 
   * If the view is not available, an error message is shown as a notification in VSCode.
   */
  async show() {
    if (this.view)
      this.view.show();
    else
      window.showErrorMessage(
        "Can't show Stack Spot view because it has been disposed or not yet started. Please, restart the extension."
      );
  }
  /**
   * Gets the current WebviewView.
   * 
   * If the vire has not yet been created, undefined is returned.
   * 
   * @returns the current view.
   */
  getView() {
    return this.view;
  }
}

class MessageHandler {
  deps;
  getStateCalls = /* @__PURE__ */ new Map();
  setStateCalls = /* @__PURE__ */ new Map();
  /**
   * If the client is offline (sendMessage returns false), the message is enqueued instead.
   * The queue is consumed as soon as the client becomes online again.
   */
  queue = [];
  streaming = /* @__PURE__ */ new Map();
  constructor(deps) {
    this.deps = deps;
    this.listen();
  }
  async sendMessageToClient(message, onNotSent) {
    const sent = await this.deps.sendMessageToClient(message);
    if (!sent) {
      if (onNotSent)
        onNotSent();
      else
        this.queue.push(message);
    }
  }
  async handleRequestToBridge(message) {
    logger.debug("handling bridge call:", message);
    const { id, property = "", arguments: args = [] } = message;
    const fn = this.deps.getBridgeHandler(property);
    if (!fn) {
      const error = `"${property}" is not a method or function of the bridge provided to the VSCodeWebview.`;
      this.sendMessageToClient(buildBridgeError(id, error));
    }
    let functionHasExecuted = false;
    try {
      const result = await fn(...args);
      functionHasExecuted = true;
      logger.debug("sending bridge response to client:", result);
      this.sendMessageToClient(buildBridgeResponse(id, result));
    } catch (error) {
      const message2 = functionHasExecuted ? `Error while sending message to client. Please make sure the return value of the method "${property}" in the Bridge provided to the VSCodeWebview is serializable.` : errorToString(error);
      this.sendMessageToClient(buildBridgeError(id, message2));
    }
  }
  handleStateResponse(type, message) {
    logger.debug(`handling ${type} state response:`, message);
    const map = type === "get" ? this.getStateCalls : this.setStateCalls;
    const promise = map.get(message.id);
    if (!promise) {
      const warning = type === "get" ? `received a state value from the client, but nothing awaits a state value in the backend. Message: ${message}.` : `finished setting a state on the client, but nothing awaits this operation in the backend. Message: ${message}`;
      return logger.warn(warning);
    }
    if (message.error) {
      return promise.reject(message.error);
    }
    promise.resolve(message.result);
    map.delete(message.id);
  }
  handleClientReadyness() {
    const q = this.queue;
    this.queue = [];
    q.forEach((m) => this.sendMessageToClient(m));
    this.streaming.forEach((value) => {
      const message = value.pending;
      const content = value.result;
      value.pending = void 0;
      value.result = "";
      if (message)
        this.stream({ ...message, content });
    });
  }
  listen() {
    this.deps.listenToMessagesFromClient(async (message) => {
      switch (message?.type) {
        case messageType.bridge:
          await this.handleRequestToBridge(message);
          break;
        case messageType.getState:
          this.handleStateResponse("get", message);
          break;
        case messageType.setState:
          this.handleStateResponse("set", message);
          break;
        case messageType.ready:
          this.handleClientReadyness();
      }
    });
  }
  dispose() {
    this.getStateCalls.forEach((value, key) => {
      value.reject(`The webview closed before the state "${String(key)}" could be retrieved.`);
    });
    this.setStateCalls.forEach((value) => {
      value.reject("The webview closed before the state could be set.");
    });
    this.getStateCalls.clear();
    this.setStateCalls.clear();
  }
  getState(name) {
    logger.debug("getting state from client:", name);
    const state = String(name);
    const message = buildGetStateRequest(state);
    let manualPromise = this.getStateCalls.get(message.id);
    if (!manualPromise) {
      manualPromise = new ManualPromise();
      this.getStateCalls.set(message.id, manualPromise);
      logger.debug("sending get state message to client:", message);
      this.sendMessageToClient(message);
    }
    return manualPromise.promise;
  }
  setState(name, value) {
    logger.debug("setting state in client:", name, value);
    const state = String(name);
    const manualPromise = new ManualPromise();
    try {
      const message = buildSetStateRequest(state, value);
      this.setStateCalls.set(message.id, manualPromise);
      logger.debug("sending set state message to client:", message);
      this.sendMessageToClient(message);
    } catch (error) {
      manualPromise.reject(
        `Can't set state with name "${state}". Please, make sure the value passed as parameter is serializable. Cause: ${error}.`
      );
    }
    return manualPromise.promise;
  }
  stream(message) {
    const currentStreaming = this.streaming.get(message.id) ?? { index: -1, result: "" };
    this.streaming.set(message.id, currentStreaming);
    currentStreaming.result += message.content;
    if (currentStreaming.pending)
      return currentStreaming.pending = message;
    currentStreaming.index++;
    const withIndex = { ...message, index: currentStreaming.index, type: "vscode-webview-stream" };
    this.sendMessageToClient(withIndex, () => {
      currentStreaming.index = -1;
      currentStreaming.pending = message;
    });
    if (message.complete || message.error)
      this.streaming.delete(message.id);
  }
}

class VSCodeWebviewBridge {
  #messageHandler;
  state;
  constructor(webview) {
    this.#messageHandler = this.#createMessageHandler(webview);
    this.state = {
      get: (key) => this.#messageHandler.getState.apply(this.#messageHandler, [String(key)]),
      set: (key, value) => this.#messageHandler.setState.apply(this.#messageHandler, [String(key), value])
    };
  }
  #createMessageHandler(webview) {
    return new MessageHandler({
      getBridgeHandler: (name) => {
        const member = this[name];
        return typeof member === "function" ? (...args) => member.apply(this, args) : void 0;
      },
      sendMessageToClient: async (message) => webview.postMessage.apply(webview, [message]),
      listenToMessagesFromClient: (listener) => {
        webview.onDidReceiveMessage((data) => {
          logger.debug("received message from client:", data);
          const message = asWebViewMessage(data);
          if (message)
            listener(message);
        });
      }
    });
  }
  /**
   * Streams a string from the extension to the webview.
   * @param message the package to send.
   */
  stream(message) {
    this.#messageHandler.stream(message);
  }
  dispose() {
    this.#messageHandler.dispose();
  }
}

export { VSCodeViewProvider, VSCodeWebview, VSCodeWebviewBridge };
//# sourceMappingURL=index.mjs.map
